/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *    Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: cbc-frame.c,v 1.14 1998/12/20 18:41:28 root Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <assert.h>

#include "common-stuff.h"
#include "peks-random.h"
#include "cbc-frame.h"
#include "messages.h"
#include "cipher.h"

#define __IOSTREAM_PRIVATE__ /* prevent from using send/recv auto-replaces */
#include "iostream.h"

#if CHAR_BIT != 8
#error Not supporting this architecture: CHAR_BIT != 8
#endif

#if CBC_IO_BUFLEN < 128 || 0x7ffff < CBC_IO_BUFLEN
# error  buffer length no in range: 128 <= CBC_IO_BUFLEN <= 0x7ffff
#endif

/* ---------------------------------------------------------------------- *
 *  the internal data type that carries all the information at run time   *
 * ---------------------------------------------------------------------- */

/* Input buffer to allow reading the total block size, all at once. */
typedef struct cipher_iocache {

  unsigned        dim ;		/* max size of cache data */
  unsigned char *data ;		/* pointer to the cache data */

  unsigned   eof ;		/* mark end of file */
  unsigned start ;		/* data [start]        is the first byte */
  unsigned  fill ;		/* date [start+fill-1] is the last byte */

} cipher_iocache ;


typedef struct ioCipher {

  cipher_desc   *cipher;	/* current cipher */
  frame_desc    *frame ;	/* current frame */

  unsigned char *block ;	/* cbc context: last chained block */
  unsigned       blen  ;	/* cbc block length */

  void (*transfer) (struct ioCipher *, 
		    unsigned char *out, const unsigned char *in);

  unsigned  cmd_pending ;	/* embedded command not executed, yet */ 

  int                fd ;	/* read/write file handle */
  unsigned     maxblock ;	/* maximal (non-fragmented) block length */
  cipher_iocache  cache ;	/* input is read and encrypted blockwise */

  unsigned serial ;		/* crc is chained with the last block */
  unsigned offset ;		/* hash values - usually 16 bytes - use this
                                   to extract from a 4 byte window */
} ioCipher ;


/* ---------------------------------------------------------------------------- *
 *                      private helpers                                         *
 * ---------------------------------------------------------------------------- */

static void
destroy_ioCipher (desc, fd)
     ioCipher   *desc ;
     unsigned    fd;
{
  if (desc->cache.data != 0)
    XFREE (desc->cache.data) ;
  if (desc->block != 0)
    XFREE (desc->block) ;
  if (desc->cipher != 0)
    destroy_cipher (desc->cipher) ;
  if (desc->frame != 0)
    destroy_frame (desc->frame) ;
}

#ifndef _HAVE_MEMMOVE
/* redefine as prototypes may hang around */
#define memmove(x,y,z) movemem (x, y, z)

static void
movemem (trg, src, len)
     unsigned char *trg;
     unsigned char *src;
     unsigned       len;
{
  unsigned n ;
  
  if (trg + len <= src ||
      src + len <= trg) {
    /* ranges do not overlap */
    if (len != 0)
      memcpy (trg, src, len) ;
    return ;
  } 

  /* shift left */
  if (trg < src) {

    /* do the non-overlapping part */
    memcpy (trg, src, n = src - trg) ;
    trg += n ;
    src += n ;
    len -= n ;
  
    /* do the rest */
    while (len --)
      * trg ++ = * src ++ ;
    return ;
  }

  /* shift right (if there is something to do, at all) */
  if (src == trg)
    return;
  
  /* do the non-overlapping part */
  n = trg - src ;
  memcpy (trg + len - n, src + len - n, n) ;
  len -= n ;

  /* do the rest */
  while (len --)
    trg [len] = src [len] ;
}
#endif /* HAVE_MEMMOVE */

#ifndef XREALLOC
static void *
xrealloc (void *p, unsigned n)
{
  void *q = realloc (p, n) ;
  if (q == 0) {
    fprintf (stderr, "Could not re-allocate to %u bytes !\n", n);
    exit (1);
  }
  return q ;
}
#define XREALLOC_DOES_NOT_INITALIZE
#define XREALLOC(p,n)	xrealloc (p, n)
#endif

/* ---------------------------------------------------------------------------- *
 *                      private input cache handling                            *
 * ---------------------------------------------------------------------------- */

static int
extract_from_io_cache (ioc, trg, len)
  cipher_iocache *ioc ;
  char           *trg ;
  unsigned        len ;
{
  /* beware of ioc == NULL */

  /* prevent from a cache underflow */
  if (ioc->fill < len)
    len = ioc->fill ;

  if (len) {
    /* move to target buffer */
    memcpy (trg, ioc->data + ioc->start, len) ;
    if ((ioc->fill -= len) == 0)
      ioc->start = 0 ;
    else
      ioc->start += len ;
  }

  return len ;
}


static void
append_to_io_cache (ioc, buf, len)
  cipher_iocache *ioc ;
  char           *buf ;
  unsigned        len ;
{
  /* beware of ioc == NULL */

  unsigned max_buf_usage = (ioc->dim >> 1) ;

  /* Appending a text [0..len-1] to the input cache data, we need to 
     check whether we must compact the input cache. This is done by
     a shift left of the text already stored. */
  
  if (max_buf_usage < ioc->start + ioc->fill + len) { 
    
    assert (ioc->fill + len <= ioc->dim) ;

    if (ioc->fill) 
      /* there is no more space for len bytes, so shift left */
      memmove (ioc->data, ioc->data + ioc->start, ioc->fill) ;
    
    /* ok, define the new start of the input cache */ 
    ioc->start = 0 ; 
  }
  
  /* append the input block to the cache */
  memcpy (ioc->data + ioc->start, buf, len) ;
  ioc->fill += len;
}


static int
resize_ioCache (ioc, new_size) 
     cipher_iocache *ioc ;
     unsigned new_size;
{
  /* prevent from a cache underflow */
  if (ioc->fill >= new_size) {
    errno = errnum (CBC_CACHERES_BLOCKED) ;
    return -1;
  }

  /* shift data to the left */
  if (ioc->start) {
    memmove (ioc->data, ioc->data + ioc->start, ioc->fill) ;
    ioc->start = 0 ;
  }

  /* adjust */
  ioc->data = XREALLOC (ioc->data, new_size) ;

  return ioc->dim = new_size ;
}


/* note, that this function has another descriptor type */
static void
flush_ioCache (desc) 
     ioCipher *desc;
{
  desc->cache.fill = 
    desc->cache.start = 0; 
}

/* ---------------------------------------------------------------------------- *
 *                      private cbc functions                                   *
 * ---------------------------------------------------------------------------- */

static unsigned
supported_block_length (n)
     unsigned n;
{
  switch (n) {
  case  8:
  case 16:
    return 1;
  }
  return 0;
}

#define p32(p)    ((u32*)p)

static void
cbc_encrypt (desc, out, in)
     ioCipher           *desc;
     unsigned       char *out;
     const unsigned char  *in;
{
  /* cbc-encrypt (E (), block [], in []) -> out []
   *
   *	block [] := E (block [] XOR in [])
   *	out   [] := block []
   */

  /* block [] := block [] XOR in [] */
  p32 (desc->block) [0] ^= p32 (in) [0] ;
  p32 (desc->block) [1] ^= p32 (in) [1] ;
  if (desc->blen > 8) {
    p32 (desc->block) [2] ^= p32 (in) [2] ;
    p32 (desc->block) [3] ^= p32 (in) [3] ;
  }
  
  /* out [] := ENCRYPT (block []) */
  XCRYPT (desc->cipher, out, desc->block);

  /* block [] := out [] */
  memcpy (desc->block, out, desc->blen);
}


static void
cbc_decrypt (desc, out, in)
     ioCipher           *desc;
     unsigned       char *out;
     const unsigned char  *in;
{
  /* cbc-decrypt (D (), block [], in []) -> out []
   *
   *	out   [] := D (in []) XOR block []
   *	block [] := in []
   */

  unsigned char buf [16] ;
  XCRYPT (desc->cipher, buf, in);

  /* out [] := D (in []) XOR block [] */
  p32 (out) [0] = (p32 (buf) [0] ^ p32 (desc->block) [0]) ;
  p32 (out) [1] = (p32 (buf) [1] ^ p32 (desc->block) [1]) ;
  if (desc->blen > 8) {
    p32 (out) [2] = (p32 (buf) [2] ^ p32 (desc->block) [2]) ;
    p32 (out) [3] = (p32 (buf) [3] ^ p32 (desc->block) [3]) ;
  }
  
  /* block [] := in [] */
  memcpy (desc->block, in, desc->blen);
}

/* ---------------------------------------------------------------------------- *
 *                      private functions: io functions                         *
 * ---------------------------------------------------------------------------- */

/*
 *  The protocol block frame looks like:
 *
 * +--+--+--+--+--+--+--+--+---+-  ... --+--+- .. --+--+--+--+--+--+
 * |  LENGTHX  |  LENGTHY  | P |  padding   |  data    |    CRC    |
 * +--+--+--+--+--+--+--+--+---+-  ... --+--+- .. --+--+--+--+--+--+
 *                              <-- P&15 -->
 *  <-------------------- LENGTHX ^ LENGTHY ---------------------->
 *  <----------------------- CRC DISTANCE ------------------------>
 *
 *  where (LENGTHX ^ LENGTHY) % 8 == 0.  The CRC is calculated with
 *  the last two bytes filled with the CRC from the previous block.
 *  This definition allows 8 and 16 bytes block size usable for 64
 *  bits and 128 bits cipher blck chaining mode.
 *
 *  All numbers are stored in big endian.
 *
 *  The value C = P & ~15 is reserved as embedded command with:
 */

#define CHANGE_SESSION_KEY 16
#define EMBD_COMMAND_MASK  16 /* masking all command bits */


#define PROTO_OVERHEAD     30 /* 13 plus at most 15 padding bytes */


static int
sendby_ioCipher_block (desc, p, cnt, flags)
     ioCipher  *desc;
     const char   *p;
     unsigned    cnt;
     unsigned  flags;
{
  /* Calculate the mask corresponding to the block size 8, or 16. So
     x % blen is equivalent to x & mask. */
  unsigned mask = (desc->blen - 1) ;

  /* Calculate the padding length so that the totel block length is a
     multiple of blen, i.e. find x so that 13 + cnt + x == 0 (mod blen),
     so x == - 13 - cnt (mod blen). Using the fact, that blen is 8, or
     16 we gain x == 19 - cnt (mod blen) */
  unsigned padding = (19 - cnt) & mask ;

  int n, len, processed ;
  unsigned char *plainbuf = ALLOCA (desc->maxblock + PROTO_OVERHEAD) ;
  unsigned char  *sendbuf = ALLOCA (desc->maxblock + PROTO_OVERHEAD) ;

  /* fill the assembly buffer key_buf with enough random data: 
     4 x random (LENGTHX), 5 x junk, 15 x padding */
  setRandomBytes (plainbuf, 26) ;

  /* Calculate the total block length rounded up to the next multiple
     of blen (the value of padding has been figured out, already.) */
  len = 13 + padding + cnt ;

  /* there first 4 bytes are random followed by the xor-part of the
     4 byte length header, note that COMMON_BUFFER_SIZE < 0x8000 */
  plainbuf [4]  = ((len >> 24) ^ plainbuf [0]) & 0xff ;
  plainbuf [5]  = ((len >> 16) ^ plainbuf [1]) & 0xff ;
  plainbuf [6]  = ((len >>  8) ^ plainbuf [2]) & 0xff ;
  plainbuf [7] &= mask ;  /* leave the last 3 or 4 random bits */
  plainbuf [7] |=  (len        ^ plainbuf [3]) & ~mask ;

  /* the least significant four bits of the 9th byte bear the
     length of the padding following, in case blen == 8 the 4th
     bit is simply unused, so leave it at random */
  plainbuf [8] &=   ~mask ;
  plainbuf [8] |= padding ;

  /* move embedded command bits to io buffer */
  plainbuf [8] &=                     ~EMBD_COMMAND_MASK  ;
  plainbuf [8] |= (desc->cmd_pending & EMBD_COMMAND_MASK) ;
  desc->cmd_pending = 0 ;

  /* add the argument text */
  memcpy (plainbuf + 9 + padding, p, cnt);

  /* add (temporary) CRC from last block */
  plainbuf [len - 4] = (desc->serial >> 24) ;
  plainbuf [len - 3] = (desc->serial >> 16) ;
  plainbuf [len - 2] = (desc->serial >>  8) ;
  plainbuf [len - 1] = (desc->serial      ) ;

  /* Calculate the crc trailer for this block while saving the new crc */
  XCRCFIRST (desc->frame, plainbuf, len);
  desc->serial = XCRCRESULT (desc->frame) ;

  /* overwrite serial count by crc trailor */
  plainbuf [len - 4] = (desc->serial >> 24) ;
  plainbuf [len - 3] = (desc->serial >> 16) ;
  plainbuf [len - 2] = (desc->serial >>  8) ;
  plainbuf [len - 1] = (desc->serial      ) ;
  
  /* encrypt the whole lot blockwise */
  for (processed = 0; processed < len; processed += desc->blen)
    (*desc->transfer) (desc, sendbuf + processed, plainbuf + processed) ;
    
  /* send that stuff */
  for (processed = 0; len > 0; processed += n, len -= n) 
    if ((n = send (desc->fd, sendbuf + processed, len, flags)) <= 0) {
      DEALLOCA (plainbuf);
      DEALLOCA  (sendbuf);
      return n;
    }
  
  DEALLOCA (plainbuf);
  DEALLOCA  (sendbuf);
  return cnt ;
}


static int
sendby_ioCipher (desc, msg, len, flags)
     ioCipher   *desc ;
     const char *msg;
     unsigned    len;
     unsigned    flags;
{
  int n, bytes = 0 ;
  
  /* need to fragment ? */
  while (len > desc->maxblock) {
    if ((n = sendby_ioCipher_block (desc, msg, desc->maxblock, flags)) < 0)
      return n ;
    bytes += n ;
    len -= desc->maxblock ;
  }
  
  if ((n = sendby_ioCipher_block (desc, msg, len, flags)) < 0)
    return n ;
  return bytes + n ;
}


static int
recfrom_ioCipher (desc, buf, cnt, flags)
     ioCipher     *desc;
     unsigned char *buf;
     unsigned       cnt;
     unsigned     flags;
{
  /* Calculate the mask corresponding to the block size 8, or 16. So
     x % blen is equivalent to x & mask. */
  unsigned mask = (desc->blen - 1) ;

  int n, len ;

  unsigned processed, padding, crc ;
  unsigned char *inbuf, *plainbuf ;

  /* check, whether there is some old stuff, left over */
  if (desc->cache.fill)
    return extract_from_io_cache (&desc->cache, buf, cnt) ;
  
  /* check whether we reached end-of-file, already */
  if (desc->cache.eof)
    return 0 ;

  inbuf    = ALLOCA (desc->maxblock + PROTO_OVERHEAD) ;
  plainbuf = ALLOCA (desc->maxblock + PROTO_OVERHEAD) ;

  /* get the first packet which has the minimum block size 16 */
  if ((n = recv (desc->fd, inbuf, 16, flags)) < 16) {
    if (n == 0) {
      errno = errnum (CBC_RECV_NULL_BLOCK) ;
      return -1 ;
    }
    desc->cache.eof = 1 ; /* mark end of file */
    DEALLOCA (plainbuf);
    DEALLOCA    (inbuf);
    return 0 ;
  }

  /* As we are on a packet boundary, we can decode the first part. Note
     that the cbc block size is either 16, or 8. In the latter case
     we need to decrypt twice */
  (*desc->transfer) (desc, plainbuf, inbuf) ;
  if (desc->blen == 8)
    (*desc->transfer) (desc, plainbuf + 8, inbuf + 8) ;

  /* read the total input block size */
  len  = ((plainbuf [0] ^ plainbuf [4]) << 24) & 0xff000000 ;
  len  = ((plainbuf [1] ^ plainbuf [5]) << 16) & 0x00ff0000 ;
  len  = ((plainbuf [2] ^ plainbuf [6]) <<  8) & 0x0000ff00 ;
  len |=  (plainbuf [3] ^ plainbuf [7])        &      ~mask ;

  if (len < 16) {
    /* total input block completely rubbish */
    errno = errnum (CBC_RECV_LEN_2SMALL) ;
    goto error_return ;
  }
  if (len > desc->maxblock) {
    /* total input block size too large */
    errno = errnum (CBC_RECV_LEN_2LARGE) ;
    goto error_return ;
  }
  if (len & mask) {
    /* len not on block boundary */
    errno = errnum (CBC_RECV_LEN_NOBNDRY) ;
    goto error_return ;
  }

  /* read the embedded command */
  desc->cmd_pending |= (plainbuf [8] & EMBD_COMMAND_MASK) ;

  /* read the padding length */
  padding = (plainbuf [8] & mask) ;

  /* read the rest of data into the temporary input buffer */
  for (processed = 16; processed < len; processed += n) {
    n = recv (desc->fd, inbuf + processed, len - processed, flags) ;
    if (n < 0) {
      /* read only some bytes (less than expected) */
      errno = errnum (CBC_RECV_UNEXP_EOF) ;
      goto error_return ;
    }
    if (n == 0) {
      desc->cache.eof = 1 ; 
      break ;
    }
  }

  /* decode that input to the plain text buffer */
  for (len = 16; len < processed; len += desc->blen)
    (*desc->transfer) (desc, plainbuf + len, inbuf + len) ;
     
  /* read crc from the input block */
  crc  = (plainbuf [len - 4] << 24) & 0xff000000 ;
  crc |= (plainbuf [len - 3] << 16) & 0x00ff0000 ;
  crc |= (plainbuf [len - 2] <<  8) & 0x0000ff00 ;
  crc |=  plainbuf [len - 1]        & 0x000000ff ;

  /* store crc from previos block and recalculate crc */ 
  plainbuf [len - 4] = (desc->serial >> 24) & 0xff ;
  plainbuf [len - 3] = (desc->serial >> 16) & 0xff ;
  plainbuf [len - 2] = (desc->serial >>  8) & 0xff ;
  plainbuf [len - 1] = (desc->serial      ) & 0xff ;

  /* compare */
  XCRCFIRST (desc->frame, plainbuf, len);
  if (crc != XCRCRESULT (desc->frame)) {
    /* read invalid crc */
    errno = errnum (CBC_RECV_CRCERR) ;
    goto error_return ;
  }

  /* save the current crc for the next block check */
  desc->serial = crc ;

  /* size of the payload */
  n = len - 13 - padding ;

  /* just a shortcut - not really necessary */
  if (desc->cache.fill == 0 && cnt >= n) {
    memcpy (buf, plainbuf + 9 + padding, n) ;
    DEALLOCA (plainbuf);
    DEALLOCA    (inbuf);
    return n ;
  }
    
  /* skip padding and append data block to io buffer */
  append_to_io_cache (&desc->cache, plainbuf + 9 + padding, n); 

  DEALLOCA (plainbuf);
  DEALLOCA    (inbuf);

  /* move input from the cache to the request buffer */
  return extract_from_io_cache (&desc->cache, buf, cnt) ;
  /*@NOTREACHED@*/
  
 error_return:
  DEALLOCA (plainbuf);
  DEALLOCA    (inbuf);
  return -1 ;
}


/* ---------------------------------------------------------------------------- *
 *                   private functions: io layer management                     *
 * ---------------------------------------------------------------------------- */

static int
resize_ioCipher_buffer (desc, new_size)
     ioCipher    *desc ;
     unsigned new_size ;
{
  if (new_size < 128 || new_size > (1024 * 128)) {
    errno = errnum (CBC_ILLRES_VALUE) ;
    return -1 ;
  }
  if (desc->cache.data != 0 && 
      resize_ioCache (&desc->cache, new_size << 1) < 0)
    return -1;
  return desc->maxblock = new_size ;
}

static int
io_control (desc, request, arg)
     ioCipher  *desc ;
     int     request ;
     void       *arg ;
{
  switch (request) {
  case IO_RESIZE_BUF:
    return resize_ioCipher_buffer (desc, *(unsigned long*)arg) ;
  case IO_CHANGE_KEY:
  }
  errno = errnum (CBC_UNKNOWN_CTL) ;
  return -1 ;
}
/* ---------------------------------------------------------------------------- *
 *                   public functions: io layer management                      *
 * ---------------------------------------------------------------------------- */

/*
 * set up an io layer implementing a particular cipher/frame
 */

int
cbc_send_init (fd, cipher, key, len)
     unsigned fd;
     const unsigned char *cipher ;
     const unsigned char *key;
     unsigned len;
{
  unsigned long start = 0;
  unsigned offset = 0;
  
  ioCipher *layer ;
  void *cclass, *fclass ;
  unsigned keylen ;

  /* initiate cipher and frame type */
  if (find_classes_by_name (&cclass, &fclass, cipher) == 0) 
    return -1 ;

  /* push a new io layer */
  if ((layer = io_push
       (fd, sizeof (ioCipher), sendby_ioCipher, io_control, 0,
	destroy_ioCipher, 1|IO_PUSH_EXCLUSIVELY)) == 0)
    return -1 ;

  if ((layer->cipher = create_encryption (cclass, key, len)) == 0 ||
      !supported_block_length (layer->cipher->blocklen)) {
    io_pop (fd, 1) ;
    return -1;
  }

  /* install cbc function initialized cbc buffer */
  layer->transfer = cbc_encrypt ;
  layer->blen     = layer->cipher->blocklen ;
  layer->block    = XMALLOC (layer->blen) ;
  
  /* if the key is long enough, use the rest to initialize crc */
  keylen = layer->cipher->keylen ;
  if (len >= keylen + sizeof (unsigned long)) {
    unsigned n = sizeof (unsigned long) ;
    while (n --) {			/* comes out as little endian */
      start <<= CHAR_BIT ;
      start  |= key [keylen + n] ;
    }
    if (len > keylen + sizeof (unsigned long)) 
      offset = key [keylen + sizeof (unsigned long)] ;
  } else 
    if (len > keylen) 
      offset = key [keylen] ;
  
  /* install frame, will never return NULL */
  layer->frame = create_frame (fclass, start, offset) ;
  
  /* save current file handle */
  layer->fd = fd ;

  /* set maximal block length */
  layer->maxblock = CBC_IO_BUFLEN ;
  
  return 0 ;
}


int
cbc_recv_init (fd, cipher, key, len)
     unsigned fd;
     const unsigned char *cipher ;
     const unsigned char *key;
     unsigned len;
{
  unsigned long start = 0;
  unsigned offset = 0;

  ioCipher *layer ;
  void *cclass, *fclass ;
  unsigned keylen ;
  
  /* initiate cipher and frame type */
  if (find_classes_by_name (&cclass, &fclass, cipher) == 0) 
    return -1 ;

  /* push a new io layer */
  if ((layer = io_push
       (fd, sizeof (ioCipher), recfrom_ioCipher, io_control, flush_ioCache, 
	destroy_ioCipher, 0|IO_PUSH_EXCLUSIVELY)) == 0)
    return -1 ;

  /* install cipher keys, will fail on key problems */
  if ((layer->cipher = create_decryption (cclass, key, len)) == 0 ||
      !supported_block_length (layer->cipher->blocklen)) {
    io_pop (fd, 0) ;
    return -1;
  }

  /* install cbc function initialized cbc buffer */
  layer->transfer = cbc_decrypt ;
  layer->blen     = layer->cipher->blocklen ;
  layer->block    = XMALLOC (layer->blen) ;
  
  /* if the key is long enough, use the rest to initialize crc */
  keylen = layer->cipher->keylen ;
  if (len >= keylen + sizeof (unsigned long)) {
    unsigned n = sizeof (unsigned long) ;
    while (n --) {			/* comes out as little endian */
      start <<= CHAR_BIT ;
      start  |= key [keylen + n] ;
    }
    if (len > keylen + sizeof (unsigned long)) 
      offset = key [keylen + sizeof (unsigned long)] ;
  } else 
    if (len > keylen) 
      offset = key [keylen] ;
  
  /* install frame, will never return NULL */
  layer->frame = create_frame (fclass, start, offset) ;
  
  /* set maximal block length */
  layer->maxblock = CBC_IO_BUFLEN ;

  /* install input cache */
  layer->cache.dim  = (layer->maxblock << 1) ;
  layer->cache.data = XMALLOC (layer->cache.dim) ;

  /* save current file handle */
  layer->fd = fd ;

  return 0 ;
}


unsigned
cbc_keylen (cipher)
     const char *cipher ;
{
  char *frame ;
  unsigned len ;

  /* split cipher name from the frame name */
  if ((frame = strchr (cipher, '/')) != 0) {
    int n = frame - cipher ;
    cipher = strncpy (ALLOCA (n+1), cipher, n) ;
    ((char*)cipher) [n] = '\0' ;
    frame ++ ;
  }

  if ((len = cipher_keylen (cipher)) != 0 && frame != 0)
    /* some more bytes to initalize the crc */
    len += sizeof (unsigned long) + 1 ;

  if (frame != 0)
    DEALLOCA (cipher) ;

  return len ;
}
