/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: iostream.c,v 1.4 1998/12/20 17:27:30 root Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *   This module sets a generic layer on top of the standard io functions
 *   like send, recv, read and write.  It replaces the channel numbers by
 *   a generic descriptor and calls the approriate io functions.
 */

#include "common-stuff.h"

#include <sys/types.h>
#include <sys/socket.h>

#define __IOSTREAM_PRIVATE__ /* prevent from using send/recv auto-replaces */
#include "iostream.h"

typedef
struct _io_desc {
  void   *desc_ptr ;
  unsigned enabled ;
  io_fn         fn ;
  destroy_fn    df ;
  ctl_fn       ctl ;
} io_desc ;


typedef
struct _io_table {
  io_desc rd ;
  io_desc wr ;
  void (*fl) (void *);
} io_table ;


static io_table *rw_table ;
static unsigned  rw_table_dim = 0 ;


/* fall back functions */
#define SEND(fd, buf, len, flg)  send (fd, buf, len, flg)
#define RECV(fd, buf, len, flg)  recv (fd, buf, len, flg)


/* ---------------------------------------------------------------------------- *
 *                      private helpers                                         *
 * ---------------------------------------------------------------------------- */

#ifndef XREALLOC
static void *
xrealloc (void *p, unsigned n)
{
  void *q = realloc (p, n) ;

  if (q == 0) {
    fprintf (stderr, "Could not re-allocate to %u bytes !\n", n);
    exit (1);
  }
  
  return q ;
}

#define XREALLOC_DOES_NOT_INITALIZE
#define XREALLOC(p,n)	xrealloc (p, n)
#endif

/* ---------------------------------------------------------------------------- *
 *                      private functions: layer management                     *
 * ---------------------------------------------------------------------------- */

static io_table *
get_io_entry (fd)
     unsigned fd ;
{

  /* make table entry fd in rw_table [] */
  if (fd >= rw_table_dim) {
    if (rw_table == 0) {
      rw_table = XMALLOC ((fd+1) * sizeof (io_table)) ;
    } else {
      rw_table = XREALLOC (rw_table, (fd+1) * sizeof (io_table)) ;
#     ifdef XREALLOC_DOES_NOT_INITALIZE
      memset (rw_table + rw_table_dim, 0, (fd+1 - rw_table_dim) * sizeof (io_table)) ;
#     endif
    }
    /* set to new dimension */
    rw_table_dim = fd + 1;
  }

  return rw_table + fd ;
}


static void
readjust_table_dim ()
{
  unsigned n = rw_table_dim ;

  /* find least index so that all entries above are unused */
  while (n --  &&
	 rw_table [n].rd.desc_ptr == 0 &&
	 rw_table [n].wr.desc_ptr == 0)
    ;
  
  /* so rw_table [n] is the first one, used */
  if (++ n < rw_table_dim) {
    if (n == 0) {
      XFREE (rw_table) ;
      rw_table = 0 ;
    } else
      rw_table = XREALLOC (rw_table, n * sizeof (io_table)) ;
    rw_table_dim = n ;
  }
}


static void *
push_any (d, size, fn, ctl, df, excl)
     io_desc *d;
     unsigned size;
     io_fn fn;
     ctl_fn ctl ;
     destroy_fn df ;
     unsigned excl;
{
  if (excl && d->desc_ptr != 0)
    return 0 ;
  d->fn  = fn ;
  d->df  = df ;
  d->ctl = ctl ;
  if (d->desc_ptr != 0)
    XFREE (d->desc_ptr) ;
  return
    d->desc_ptr = XMALLOC (size) ;
}


static void
pop_any (d, fd)
     io_desc *d;
     unsigned fd ;
{
  if (d->desc_ptr != 0) {
    if (d->df != 0)
      (*d->df) (d->desc_ptr, fd) ;
    XFREE (d->desc_ptr) ;
  }
  memset (d, 0, sizeof (io_desc)) ;
  readjust_table_dim () ;
}


/* ---------------------------------------------------------------------------- *
 *                      public functions: layer management                      *
 * ---------------------------------------------------------------------------- */

void *
io_push (fd, size, fn, ctl, fl, df, how)
     unsigned fd;
     unsigned size;
     io_fn fn ;
     ctl_fn ctl ;
     void (*fl) (void *);
     destroy_fn df ;
     unsigned how ;
{
  unsigned excl = how & IO_PUSH_EXCLUSIVELY ;
  io_table *t = get_io_entry (fd) ;

  switch (how & 3) {
  case 0:
    t->fl = fl ;
    return push_any (& t->rd, size, fn, ctl, df, excl) ;
  case 1:
    return push_any (& t->wr, size, fn, ctl, df, excl) ;
  }
  
  return 0 ;
}



void
io_pop (fd, how)
     unsigned fd;
     unsigned how;
{
  if (fd < rw_table_dim) 
    switch (how) {
    case 2:
    case 0:
      pop_any (& rw_table [fd].rd, fd) ;
      if (!how)
	return ;
    case 1:
      pop_any (& rw_table [fd].wr, fd) ;
    }
}


int 
io_shutdown (fd, how)
     unsigned fd;
     unsigned how;
{
  io_pop (fd, how);
  return shutdown (fd, how);
}

int 
io_ctrl (fd, req, arg, how)
     unsigned fd;
     int     req; 
     void   *arg; 
     unsigned how; 
{
  void *desc;
  ctl_fn fn ;

  if (fd < rw_table_dim) 
    switch (how) {
    case 0:
      if ((desc = rw_table [fd].rd.desc_ptr) == 0)
	return -1 ;
      if ((fn = rw_table [fd].rd.ctl) == 0)
	return -1 ;
      return (*fn) (desc, req, arg) ;
    case 1:
      if ((desc = rw_table [fd].wr.desc_ptr) == 0)
	return -1 ;
      if ((fn = rw_table [fd].wr.ctl) == 0)
	return -1 ;
      return (*fn) (desc, req, arg) ;
    }
    
  return -1 ;
}

void
io_enable (fd, how)
     unsigned fd;
     unsigned how;
{
  if (fd < rw_table_dim) 
    switch (how) {
    case 2:
    case 0:
      rw_table [fd].rd.enabled = 1 ;
      if (!how)
	return ;
    case 1:
      rw_table [fd].wr.enabled = 1 ;
    }
}


void
io_suspend (fd, how)
     unsigned fd;
     unsigned how;
{
  if (fd < rw_table_dim) 
    switch (how) {
    case 2:
    case 0:
      rw_table [fd].rd.enabled = 0 ;
      if (!how)
	return ;
    case 1:
      rw_table [fd].wr.enabled = 0 ;
    }
}


/* ---------------------------------------------------------------------------- *
 *                      public functions: io functions                          *
 * ---------------------------------------------------------------------------- */

int
io_send (fd, buf, len, flags)
     unsigned fd;
     const char *buf;
     unsigned len;
     unsigned flags;
{
  void *desc ;
  io_fn fn;
  
  if (fd >= rw_table_dim ||
      rw_table [fd].wr.enabled == 0 ||
      (desc = rw_table [fd].wr.desc_ptr) == 0 ||
      (fn   = rw_table [fd].wr.fn)       == 0)
    
    return SEND (fd, buf, len, flags) ;
  
  return (*fn) (desc, buf, len, flags);
}


int
io_recv (fd, buf, len, flags)
     unsigned    fd;
     char      *buf;
     unsigned   len;
     unsigned flags;
{
  void *desc ;
  io_fn fn;
  
  if (fd >= rw_table_dim ||
      rw_table [fd].rd.enabled == 0 ||
      (desc = rw_table [fd].rd.desc_ptr) == 0 ||
      (fn   = rw_table [fd].rd.fn)       == 0)
    
    return RECV (fd, buf, len, flags) ;
      
  return (*fn) (desc, buf, len, flags);
}


void
rd_flush (fd)
     unsigned fd;
{
  void *desc, (*fl)(void *) ;
  
  if (fd >= rw_table_dim ||
      (fl   = rw_table [fd].fl) == 0 ||
      (desc = rw_table [fd].rd.desc_ptr) == 0)
    
    return ;

  (*fl) (desc) ;
}
