/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-calc.c,v 1.2 1998/12/14 21:26:53 root Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include <limits.h>

#include "common-stuff.h"
#include "peks-internal.h"
#include "messages.h"
#include "peks-random.h"

#undef DEBUG_CALC_GENERATE_MODULES

/* ---------------------------------------------------------------------------- *
 *                private functions: prime and other tests                      *
 * ---------------------------------------------------------------------------- */

static int
is_not_miller_rabin_prime (p, num_tests)
     int num_tests;
     mpz_t p;
{
  /* References: 
   *
   *  - Bruce Schneier; Applied Cryprography; NY, Wiley & Sons, 2nd ed 1996, 
   *    chapter 11.5, pg 260 ff.
   *  - D.R. Stinson; Cryptography; Boca Raton, CRC Press 1995, chapter 4,
   *    Fig 4.8 pg 137 ff. 
   */
  
  mpz_t p_min_1, m, a, z;
  unsigned b, is_definitely_composite;

  unsigned rlen = 32 ;			/* max size of a random number */

  mpz_init (a);				/* will a (small) random a < p */

  if ((b = mpz_sizeinbase (p, 2) - 1) < rlen)
    rlen = b ;				/* length for a randum mumber */

  mpz_init (p_min_1) ;
  mpz_sub_ui (p_min_1, p, 1) ;
  b = mpz_scan1 (p_min_1, 0) ;		/* max b with 2^b | p-1 */
  
  mpz_init (m);
  mpz_tdiv_q_2exp (m, p_min_1, b);	/* so p-1 = m * 2^b */

  mpz_init (z);
  do					/* do this test at least once */
    {
      unsigned j ;
      
      is_definitely_composite = 1;	/* default assumption: no prime */
      getMpzRandomNumber (&a,rlen);	/* get new random number */

      mpz_powm (z, a, m, p) ;		/* z := a^m mod p */

      if (mpz_cmp_ui (z, 1) == 0)	/* z == 1 mod p */
	goto p_may_be_prime ;
      
      for (j = 0; j < b; j ++)		/* see p-1 == m * 2^b */
	{
	  if (mpz_cmp (z, p_min_1) ==0)	/* z == -1 mod p */
	    goto p_may_be_prime ;
	  
	  mpz_powm_ui (z, z, 2, p) ;	/* z := z^2 mod p */
	  
	  if (mpz_cmp_ui (z, 1) == 0)	/* if z == 1 here, then also for rest */
	    break ;			/* definitely not prime */
	}
      goto clean_up_and_return ;	/* not a prime when this loop is done */
      /*@NOTREACHED@*/

    p_may_be_prime:
      is_definitely_composite = 0;	/* may be a prime, next test or done */
    } 
  while ( -- num_tests > 0) ;		/* repeat that test */
  
 clean_up_and_return:

  mpz_clear (p_min_1);
  mpz_clear       (m);
  mpz_clear       (a);
  mpz_clear       (z);

  return is_definitely_composite ;
}



static int
number_is_a_prime (num, prob_weight)
     mpz_t num;
     unsigned prob_weight;
{
  extern unsigned short small_prime_numbers [] ;
  unsigned p, r, n = 0 ;

  mpz_t OP;
  mpz_init (OP);
  
  /* check for divisibility by some small primes, first */
  while ((p = small_prime_numbers [n ++]) &&
	 /* does p | num <=> num == 0(mod p) ? */
	 (r = mpz_mod_ui (OP, num, p)))
    ;

  if (r) /* Rabin-Miller test */
    r = (is_not_miller_rabin_prime (num, prob_weight) == 0) ;
  
#ifdef _seems_not_better_than_the_miller_rabin_test_
  if (r) /* some probabilistic test due to Knuth */
    r = mpz_probab_prime_p (num, prob_weight) ;
#endif

  mpz_clear (OP);
  return r;
}


/* ---------------------------------------------------------------------------- *
 *                private functions: make primes                                *
 * ---------------------------------------------------------------------------- */

static int
set_modulus_and_generator (k)
     peks_key *k;
{
  extern unsigned short small_prime_numbers [] ;
  int n, m;				/* indices in small_prime_numbers [] */

  /* I need s list of small primes starting at 2. As the array in 
     small_prime_numbers [] starts with 3, I provide a macro SP which
     fixes that problem without changing the source smallprimes.c */
# define SP(n)  ((n) < 0 ? 2 : small_prime_numbers [n])

  unsigned r, s, t;			/* t = rs, for small primes r, s */

  unsigned g ;				/* generator (mod p) to be fonud */
  mpz_t G, OP;				/* G := (mpz_t) g */
  mpz_init (G) ;
  mpz_init (OP) ;

  /* loop over r in (2, smallprimes ...) */
  for (n = -2, r = 1; r; n++, r = SP (n))
    
    /* loop over s >= r, r * s != 1, s in (2, smallprimes ...) */
    for (m = n, s = r; s; m++, s = SP (m)) {
    
      if ((t = r * s) == 1)
	continue ;

      /* print keep alive message */
      if (k->output != 0 && (m+n) % 77 == 0) (* k->output) (".") ;

      mpz_mul_ui (OP, k->private, t);		/* OP := q * t */
      mpz_add_ui (k->modulus, OP, 1);		/* p := q * t + 1 */

      /* retry if this number is not prime */
      if (!number_is_a_prime (k->modulus, k->prime_prob, OP))
	continue ;

      /* print found message */
      if (k->output != 0) (* k->output) ("p") ;

      /* find generator g (mod p) */
      for (g = 2; g < UINT_MAX; g ++) {

	mpz_set_ui (G, g);

	/* print keep alive message */
	if (k->output != 0 && g % 11 == 0) (* k->output) (".") ;

	mpz_powm_ui (OP, G, t, k->modulus) ;		/* OP := g^t (mod p) */
	if (mpz_cmp_ui (OP, 1) == 0)			/* g^t == 1 (mod p) ? */
	  continue ;

	/* note that s >= r and at most one of r or s is 1, so s > 1 */

	if (r != 1) {
	  /* this case saves one multiplication operation */
	  mpz_powm (OP, G, k->private, k->modulus) ;
	} else {					/* means s >= r > 1 */
	  mpz_mul_ui (OP, k->private, r) ;		/* OP := q * r */
	  mpz_powm (OP, G, OP, k->modulus) ;		/* OP := g^qr (mod p) */
	}
	if (mpz_cmp_ui (OP, 1) == 0)			/* g^qr == 1 (mod p) ? */
	  continue ;
	
	mpz_mul_ui (OP, k->private, s) ;		/* OP := q * s */
	mpz_powm (OP, G, OP, k->modulus) ;		/* OP := g^qs (mod p) */
	if (mpz_cmp_ui (OP, 1) == 0)			/* g^qr == 1 (mod p) ? */
	  continue ;

	/* now, g^x != 1 (mod p) for all d | (p-1), so p is generator */
	k->generator = g ;				/* implicit return value */
	n = 1 ;						/* return code: 1 */
	goto clean_up_and_return_n;
      }
    }
  n = 0 ;						/* not found: r-code 0 */

 clean_up_and_return_n:
  if (k->output != 0) (* k->output) (n ? "g" : "!") ;
  mpz_clear (G);
  mpz_clear (OP);
  return n;
# undef SP
}


static int
set_random_prime_factor (k, bits, max_tries)
     peks_key *k;
     unsigned bits;
     unsigned max_tries;
{
  int run = max_tries;

  /* random number must be at least 'bits' large, at most 'max' */
  int max = (bits << 3) / 7;

  do							/* do at least once */
    {
      /* get fresh randum number */
      getMpzRandomNumber (&k->private, max-2);
      
      mpz_mul_2exp (k->private, k->private, 1);		/* num <<= 1 */
      mpz_setbit   (k->private, 0) ;			/* num |= 1 makes it odd */

      /* print keep alive message */
      if (k->output != 0 && (run % 33) == 0) (* k->output) (".") ;
      
      /* don't let the random number become too small */
      if (mpz_sizeinbase (k->private, 2) < bits)	/* set the msb to 1 */
	mpz_setbit (k->private, bits) ;			/* num |= 1 makes it big */
      
      /* retry if this number is not prime */
      if (!number_is_a_prime (k->private, k->prime_prob))
	continue ;

      /* print found message, return */
      if (k->output != 0) (* k->output) ("q") ;
      return (max_tries - run) ;
    }
  while (-- run > 0) ;

  if (k->output != 0) (* k->output) ("!") ;
  return 0;
}


/* ---------------------------------------------------------------------------- *
 *              global functions: fiddling around with numbers                  *
 * ---------------------------------------------------------------------------- */


int
peks_generate_modulus (key, bits, try)
     peks_key *key;
     unsigned bits;
     unsigned  try;
{
  int run = try ;
  
  if (key->output != 0) {
#   ifdef DEBUG_CALC_GENERATE_MODULES
    char buf [100] ;
    sprintf (buf, "Generating %u-bit primes: ", bits, try);
    (*key->output) (buf) ;
#   else
    (*key->output) ("Generating primes: ") ;
#   endif
  }

  if (bits == 0)
    bits = DEFAULT_PRIME_BITS ;

  if (bits > 3) 
    while (run -- > 0) {

      if (set_random_prime_factor (key, bits-2, try) == 0 ||
	  set_modulus_and_generator (key) == 0)
	continue ;
     
      if (key->output != 0) (*key->output) ("\n") ; /* print */
      key->errnum = errnum (NO_PEKS_ERROR) ;
      return (try - run);
      /*@NOTREACHED@*/
    }

  key->errnum = errnum (TOO_MANY_PRIME_TRIES) ;
  return 0 ;
}
  

