/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-file.c,v 1.2 1998/12/14 21:27:07 root Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <ctype.h>
#include <errno.h>

#include "common-stuff.h"
#include "peks-internal.h"
#include "messages.h"

/* the maximal line length in a key file */
#define MAX_LINE_LENGTH (5 * 1024)


/* ---------------------------------------------------------------------------- *
 *                  public key file i/o functions                               *
 * ---------------------------------------------------------------------------- */

/*
 * parse a file linewise to find
 *
 *	<tag>: <key defs> ...
 *
 * and load that line into a key structure.
 */

peks_key *
read_peks_key (tag, cmp, file)
     const char *tag;
     strncmp_fn cmp;
     const char *file;
{
  FILE *inp;
  unsigned len, n ;
  char read_buf [MAX_LINE_LENGTH], *s, *p ;

  /* open input file if it exists */
  if (tag == 0 || file == 0 || cmp == 0 ||
      (inp = fopen (file, "r")) == 0) 
    return 0;

  len = strlen (tag);

  /* parse input file up until the tag line */
  while (fgets (read_buf, sizeof (read_buf), inp)) {
	   
    /* skip leading spaces */
    for (s = read_buf, n = strlen (read_buf); 
	 isspace (*s) && n > 0 ; 
	 s ++, n--) 
      ;
      
    /* find end of tag */
    if ((p = strchr (s, ':')) != 0)
      len = p - s ;
    else
      len = 0 ;

    /* stop at the pattern |^<tag>: | */
    if (len &&
	s [len    ] == ':' &&
	s [len + 1] == ' ' &&
	(* cmp) (tag, s, len) == 0) {

      fclose (inp) ;
      return get_peks_key_from_str (s + len + 2) ;
    }
  }
    
  /* did not find that tag */
  fclose (inp);
  return 0 ;
}



int
save_peks_key (key, tag, cmp, file)
     peks_key *key;
     const char *tag;
     strncmp_fn cmp;
     const char *file;
{
  FILE *inp, *outp = 0;
  char read_buf [MAX_LINE_LENGTH], *new, *old, *s, *p ;
  unsigned len, n, r_code = 0 ;

  if (key == 0 || tag == 0 || file == 0)
    return 0 ;

  /* open input file if it exists */
  if ((inp = fopen (file, "r")) == 0 && errno != ENOENT) {
    key->errnum = errnum (CANT_OPEN_KEY_RFILE) ;
    return 0;
  }

  /* create output file names for "new" and "old" */
  len = strlen (file) ;
  new = ALLOCA (len + 4 + 1) ;
  strcpy (new, file);
  strcat (new, ".new") ;
  old = ALLOCA (len + 4 + 1) ;
  strcpy (old, file);
  strcat (old, ".old") ;

  /* open & truncate output file */
  if ((outp = fopen (new, "w+")) == 0) {
    key->errnum = errnum (CANT_OPEN_KEY_WFILE) ;
    goto get_ready_for_return;
  }

  len = strlen (tag);
  if (inp != 0) {

    /* parse input file up until the tag line */
    while (fgets (read_buf, sizeof (read_buf), inp)) {
      
      /* skip leading spaces */
      for (s = read_buf, n = strlen (read_buf); 
	   isspace (*s) && n > 0 ; 
	   s ++, n--) 
	;

      /* find end of tag */
      if ((p = strchr (s, ':')) != 0)
	len = p - s ;
      else
	len = 0 ;

      /* stop at the pattern |^<tag>: | */
      if (cmp != 0 && len &&
	  s [len    ] == ':' &&
	  s [len + 1] == ' ' &&
	  (* cmp) (tag, s, len) == 0) {
	break ;
      }
      
      /* check for input line overflow */
      if (s [n-1] != '\n' &&
	  s [n-1] != '\r') {
	key->errnum = errnum (KEYFILE_PARSER_ERR) ;
	goto get_ready_for_return;
      }

      /* copy to the output file */
      fputs (read_buf, inp) ;

    }
  } /* if (inp != 0) */ 
  

  /* make string from private key, make whole keys line */
  p = mpz2base64 (&key->private) ;
  s = make_peks_key_line (tag, key, p, "\n") ;

  /* clean up after writing the keys string to file */
  fputs (s, outp);
  XFREE (p);
  XFREE (s);

  if (inp != 0) {

    /* dump the rest onto file */
    while (fgets (read_buf, sizeof (read_buf), inp)) {
      
      /* check for input line overflow */
      n = strlen (read_buf) ;
      if (read_buf [n-1] != '\n' &&
	  read_buf [n-1] != '\r') {
	key->errnum = errnum (KEYFILE_PARSER_ERR) ;
	goto get_ready_for_return;
      }

      /* copy to the output file */
      fputs (read_buf, inp) ;
    }
  
    fclose (inp);
  } /* if (inp != 0) */ 

  fclose (outp);

  /* files have been closed, now */
  inp = outp = 0 ;

  /* save current key file */
  if (inp) {
    unlink (old) ;
    rename (file, old) ;
  }

  /* install new key file */
  r_code = (rename (new, file) == 0) ;
  key->errnum = errnum (NO_PEKS_ERROR) ;

 get_ready_for_return:
  
  if (inp != 0)  fclose (inp);
  if (outp != 0) fclose (outp);
  DEALLOCA (new) ;
  DEALLOCA (old) ;
  return r_code ;
}
