/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *       Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-server.c,v 1.6 1998/12/16 21:25:11 root Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"
#include "peks-internal.h"
#include "peks-random.h"
#include "messages.h"
#include "cbc-frame.h"
#include "baseXX.h"

#include <sys/types.h>
#include <sys/socket.h>

#define __IOSTREAM_PRIVATE__ /* prevent from using send/recv auto-replaces */
#include "iostream.h"

/* ---------------------------------------------------------------------------- *
 *                  private key handshake functions                             *
 * ---------------------------------------------------------------------------- */

static const char *
peks_server_setup (K, host, file, template)
     peks_key **K;
     const char *host;
     const char *file;
     peks_key *template;
{
  /* we use a jump table, here */
  
  if (host != 0 && file != 0 &&
      (*K = read_peks_key (host, hostcmp, file)) != 0)
    goto return_public_key ;
  
  /* get default key: dup_peks_key () always returns a value */
  if (template != 0)  {
    *K = dup_peks_key (template) ;
    goto return_public_key ;
  }

  if ((*K = new_peks_key (0)) != 0)
    goto return_public_key ;

  return 0 ;
  /*@NOTREACHED@*/

 return_public_key:
  return make_public_key_str (*K, 0) ;
}




static char *
peks_server_decode_sender_key (k, N, T, in)
     peks_key *k;
     unsigned *N;
     const char **T;
     const char *in;
{
  char *s, *sk, *sl;
  unsigned n, dummy ;

  if (N == 0)
    N = &dummy ;

  /* in case of an error, k->errnum will be set */
  if ((*N = accept_response_key_str (k, in)) == 0)
    goto set_errcode_and_exit;
  
  if (k->import_type != elg_encrypted) {
    k->errnum = errnum (DECODE_NO_CIPHER) ;
    goto set_errcode_and_exit;
  }

  /* extract session key s of length n */
  s = base64toBin (k->import_str, &n) ;
  
  /* 
   * parse the session key details:
   *
   *	key: <base64-key-bits> <cipher-type>
   */
  s [n-1] = '\0' ;
  if ((sk = strtok (s, PEKS_FS)) == 0 ||
      strcmp (sk, "key:")        != 0 ||
      (sk = strtok (0, PEKS_FS)) == 0 ||
      (sl = strtok (0, PEKS_FS)) == 0) {
    XFREE (s);
    k->errnum = errnum (DECODE_CLNTKEY_ERR) ;
    goto set_errcode_and_exit;
  }
  sk = base64toBin (sk, N) ;

  if (T != 0)
    *T = XSTRDUP (sl) ;
  
  XFREE (s);
  return sk;
  /*@NOTREACHED@*/

 set_errcode_and_exit:
  *N = k->errnum ;
  return 0 ;
}




static char *
peks_server_wrap_receiver_key (s, len, type)
     const char *s;
     unsigned len;
     const char *type;
{
  char *t, *sk ;

  if (len == 0)
    len = strlen (s) ;
  if (type == 0)
    type = "." ;

  /* make reciever session key as
   *
   *	key: <base64-key-bits> <cipher-type>
   */
  sk = bin2base64 (s, len) ;
  t  = XMALLOC (8 + strlen (sk) + strlen (type));
  sprintf (t, "key: %s %s\n", sk, type) ;
  XFREE (sk) ;

  return t;
}


/* ---------------------------------------------------------------------------- *
 *                  public key server handshake function                        *
 * ---------------------------------------------------------------------------- */

int
server_negotiate_session_key (key, socket, client, keyfile)
     peks_key       * key ;
     int           socket ;
     const char  * client ;
     const char * keyfile ;
{
  const char *cipher, *t ;
  char *s, *buf;
  int l, n;

  /* set up the public peks server key */
  if ((t = peks_server_setup (&key, client, keyfile, key)) == 0) 
    return errnum (SRV_SETUP_SKEY) ;

  /* send the public peks key to the client */
  send (socket, t, strlen (t), 0);
  XFREE (t);

  /* receive the peks wrapped answer from the client */
  buf = ALLOCA (CBC_IO_BUFLEN+1) ;
  if ((n = recv (socket, buf, CBC_IO_BUFLEN, 0)) < 0) {
    n = errno ;
    goto error_code_return ;
  }
  buf [n] = '\0' ;

  /* unpack the session key from the peks wrapper */
  s = peks_server_decode_sender_key (key, &l, &cipher, buf) ;
  n = key->errnum ;
  end_peks_key (key) ;
  if (s == 0)
    goto error_code_return ;
 
  /* install it as sender session key s of length l and type t */
  n = cbc_send_init (socket, cipher, s, l) ;
  XFREE (s);
  if (n != 0) {
    n = errnum (SRV_INSTALL_CKEY) ;
    goto error_code_return ;
  }
  io_enable (socket, 1) ;
  
  /* generate reciever session key s of length l */
  s = ALLOCA (l = cbc_keylen (cipher)) ;
  setRandomBytes (s, l) ;

  /* install receiver session key s of length n and type t */
  if ((n = cbc_recv_init (socket, cipher, s, l)) != 0) {
    XFREE (cipher);
    DEALLOCA (s);
    n = errnum (SRV_INSTALL_SKEY) ;
    goto pop_error_code_return ;
  }
  io_enable (socket, 0) ;

  /* peks_server_wrap_receiver_key () always returns a value, send it */
  buf = peks_server_wrap_receiver_key (s, l, cipher) ;
  XFREE (cipher);
  DEALLOCA (s);
  if (io_send (socket, buf, strlen (buf), 0) < 0) {
    n = errnum (SRV_SEND_FAILED) ;
    goto pop_error_code_return ;
  }

  /* check magic string */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN, 0)) < 0) {
    n = errnum (SRV_MAGIC_FAILED) ;
    goto pop_error_code_return ;
  }
  buf [n] = '\0' ;
  if (strcmp (buf, PEKS_MAGIC) != 0) {
    n = errnum (SRV_WRONG_MAGIC) ;
    goto pop_error_code_return ;
  }
  
  /* done */
  DEALLOCA (buf) ;
  return 0;

  /*@NOTREACHED@*/ 

 pop_error_code_return:
  io_pop (socket, 1) ;

 error_code_return:
  DEALLOCA (buf) ;
  return n;
}

