/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1998
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *    Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-setup.c,v 1.1 1998/12/13 21:45:54 root Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"
#include "peks-internal.h"
#include "peks-random.h"
#include "cipher.h"


#define CRC_TYPE  "crc16"
#define CRC_OFFS       0

/* ---------------------------------------------------------------------------- *
 *               private functions: i/o call back                               *
 * ---------------------------------------------------------------------------- */

static void
xprint (stream, text)
     FILE *stream;
     char *text;
{
  fputs (text, stream);
  fflush (stream) ;
}

static void
xprint2 (text)
     char *text;
{
  xprint (stderr, text);
}


/* ---------------------------------------------------------------------------- *
 *                  public functions: utilities                                 *
 * ---------------------------------------------------------------------------- */

unsigned
sequenced_md (x1, x2, x3)
     const char *x1, *x2, *x3;
{
  static frame_desc *frame ;
 
  if (frame == 0)
    frame = create_frame (find_frame_class (CRC_TYPE), 0, CRC_OFFS);
  
  XCRCFIRST (frame, x1, strlen (x1)) ;
  XCRCNEXT  (frame, " ", 1) ;
  XCRCNEXT  (frame, x2, strlen (x2)) ;
  if (x3 != 0) {
    XCRCNEXT  (frame, " ", 1) ;
    XCRCNEXT  (frame, x3, strlen (x3)) ;
  }
  return XCRCRESULT (frame) ;
}


/*
 * For a given list of strings [] = {<ident>, ... } and two non negative
 * numbers <major> < 100, <minor> < 100 parse a string 
 *
 *	<ident> "/" <major> "." <minor>
 *
 * and return the unique identifier
 *
 *	(index {<ident>} + 1) * 10000 + <major> * 100	+ <minor>
 *
 * if such an index exists, and 0 otherwise.
 */

unsigned
peks_split_ident (ident_table, text, len)
     const char *ident_table []; /* NULL terminated array */
     const char *text;
     unsigned len;
{
  char *t, *s ;
  int maj, min, n = 0, result = 0;

  /* get a writable copy of the argument string */
  if (len == 0) 
    len = strlen (text) ;
  (s = strncpy (ALLOCA (len + 1), text, len + 1)) [len] = '\0' ;

  if ((t = strtok (s, "/")) != 0)
    
    while (ident_table [n] != 0) {
      
      /* get a match in the lookup table */
      if (strcmp (ident_table [n ++], t) != 0)
	continue ;

      /* ok, get the rest */
      if ((t = strtok (0, PEKS_FS)) != 0 &&
	  sscanf (t, "%u.%u", &maj, &min) == 2 &&
	  maj < 100 && min < 100)
	result = (n * 100 + maj) * 100 + min ;
      break;
    }
  
  DEALLOCA (s) ;
  return result ;
}


char *
make_peks_key_line (tag, key, key_str, term)
     const char *tag;
     const peks_key *key;
     const char *key_str ;
     const char *term;
{
  unsigned e, f, a, b, c, d ;
  char *crc_str, *s, *t ;

  /* calculate crc for this line */
  crc_str = uint2base64 
    (sequenced_md (key->modulus_str, key->generator_str, key_str));

  /* there may be a prefix and a postfix */
  e =  (tag == 0) ? 0 : strlen (tag) ;
  f = (term == 0) ? 0 : strlen (term) ;
  
  t = s = XMALLOC ( e + f + 2 +
		    (a = strlen (key->modulus_str)) + 1 +
		    (b = strlen (key->generator_str)) + 1 +
		    (c = strlen (key_str)) + 1 +
		    (d = strlen (crc_str)) + 1) ;
  
  if (tag != 0) {
    strcpy (s, tag)              ; s += e ;
    strcpy (s, ": ")             ; s += 2 ;
  }
  strcpy (s, key->modulus_str)   ; s += a ;
  strcpy (s, " ")                ; s ++ ;
  strcpy (s, key->generator_str) ; s += b ;
  strcpy (s, " ")                ; s ++ ;
  strcpy (s, key_str)            ; s += c ;
  strcpy (s, " ")                ; s ++ ;
  strcpy (s, crc_str) ;          ; s += d ;
  if (term != 0)	/* no space delimiter, here */
    strcpy (s, term) ;

  XFREE (crc_str);

  return t ;
}

/* ---------------------------------------------------------------------------- *
 *                     public initialization functions                          *
 * ---------------------------------------------------------------------------- */

peks_key *
dup_peks_key (old)
     const peks_key *old;
{
  peks_key *new = XMALLOC (sizeof (peks_key));
  
  mpz_init_set (new->modulus, old->modulus);
  mpz_init_set (new->private, old->private);
  mpz_init_set (new->common,  old->common);

  new->generator  = old->generator ;
  new->prime_prob = old->prime_prob ;
  new->output     = old->output ;
  
  if (old->modulus_str   != 0)
    new->modulus_str =   XSTRDUP (old->modulus_str);
  if (old->generator_str != 0)
    new->generator_str = XSTRDUP (old->generator_str);
  if (old->export_key_str != 0) 
    new->export_key_str = XSTRDUP (old->export_key_str) ;
  if (old->import_str     != 0)
    new->import_str =     XSTRDUP (old->import_str);

  return new;
}


peks_key *
new_peks_key (bits)
     unsigned bits;
{
  peks_key *k = XMALLOC (sizeof (peks_key));
  mpz_t G, PUB ;

  k->output     = xprint2 ;
  k->prime_prob = PRIME_PROBABBILTY_WEIGHT;

  mpz_init (k->modulus) ;
  mpz_init (k->private) ;
  mpz_init (k->common) ;

  if (!peks_generate_modulus (k, bits, TRY_RANDOM_IS_PEKS_PRIME)) {    
    end_peks_key (k) ;
    return 0;
  }
  
  getMpzRandomNumber (&k->private, bits >> 2);	/* prv key a, half as big as p */

  /* calculate public key */
  mpz_init (PUB);
  mpz_init_set_ui (G, k->generator) ;
  mpz_powm (PUB, G, k->private, k->modulus);       	/* pub key := g^a (mod p) */
 
  /* make printables, re-assign strings */
  k->modulus_str = mpz2base64 (&k->modulus) ;		/* prime modulus p */
  k->generator_str = uint2base64 (k->generator) ;	/* generator g (mod p) */
  k->export_key_str = mpz2base64 (&PUB) ;		/* public key g^a (mod p) */

  /* clean up */
  mpz_clear (PUB) ;
  mpz_clear (G) ;
  return k;
}

/*
 * parse a line
 *
 *	<p> <g> <a> <crc>
 *
 * and initialize a peks key structure 
 */

peks_key *
get_peks_key_from_str (l)
     const char *l;
{
  peks_key *key = 0;
  char *line = ALLOCASTRDUP (l) ;

  char *prv_str, *crc_str, *s ;
  mpz_t PUB, G;

  /* create key structure */
  key = XMALLOC (sizeof (peks_key));

  /* parse line entries: get modulus */
  if ((key->modulus_str = strtok (line, PEKS_FS)) == 0)
    goto error_return;
  
  /* parse line entries: get generator */
  if ((key->generator_str = strtok (0, PEKS_FS)) == 0)
    goto error_return;
  
  /* parse line entries: get private key */
  if ((prv_str = strtok (0, PEKS_FS)) == 0)
    goto error_return;

  /* parse line entries: get crc */
  if ((crc_str = strtok (0, PEKS_FS)) == 0)
    goto error_return;

  /* no more entries ! */
  if ((s = strtok (0, PEKS_FS)) != 0)
    goto error_return;
    
  /* verify line entries */
    if (base64toUint (crc_str) != 
	sequenced_md (key->modulus_str, key->generator_str, prv_str))
    goto error_return;

  /* convert to internal number representation */
  mpz_init (key->modulus) ;
  mpz_init (key->private) ;
  if (base64toMpz (&key->modulus, key->modulus_str) == 0)
    goto mpz_error_return;
  if (base64toMpz (&key->private, prv_str) == 0)
    goto mpz_error_return;
  if ((key->generator = base64toUint (key->generator_str)) == UINT_MAX)
    goto mpz_error_return;
  
  /* accept line entries, allocate string spaces */
  key->modulus_str   = XSTRDUP (key->modulus_str) ;
  key->generator_str = XSTRDUP (key->generator_str) ;

  /* calculate public key, set export string */
  mpz_init (PUB);
  mpz_init_set_ui (G, key->generator) ;
  mpz_powm (PUB, G, key->private, key->modulus);	/* pub key := g^a (mod p) */
  key->export_key_str = mpz2base64 (&PUB) ;		/* public key g^a (mod p) */
  mpz_clear (PUB) ;
  mpz_clear (G) ;

  return key;
  /*@NOTREACHED@*/

 mpz_error_return:

  mpz_clear (key->modulus) ;
  mpz_clear (key->private) ;

 error_return:
  
  XFREE    (key) ;
  DEALLOCA (line) ;
  return 0 ;
}

/* ---------------------------------------------------------------------------- *
 *                     public destroy function                                  *
 * ---------------------------------------------------------------------------- */

void
end_peks_key (key)
     peks_key *key;
{
  if (key == 0)
    return ;

  mpz_clear (key->modulus) ;
  mpz_clear (key->private) ;
  mpz_clear (key->common) ;

  if (key->modulus_str    != 0)
    XFREE (key->modulus_str) ;
  if (key->generator_str  != 0)
    XFREE (key->generator_str) ;
  if (key->export_key_str != 0) 
    XFREE (key->export_key_str);
  if (key->import_str != 0)
    XFREE (key->import_str) ;
  
  memset (key, 0, sizeof (*key)) ;
  XFREE (key) ;
}


